import os
import json

from yaml import YAMLObject, Loader, load
from yaml.nodes import ScalarNode, SequenceNode, MappingNode
from .exceptions import FileParsingException


class ContextPassingYamlLoader(Loader):
    def __init__(self, stream, context):
        super().__init__(stream)
        self.context = context


class ArgListAcceptingYamlTag(YAMLObject):
    @classmethod
    def __get_exception(cls, message):
        return FileParsingException("Error constructing {tag} tag. {message}".format(
            tag=cls.yaml_tag,
            message=message,
        ))

    @classmethod
    def handle_tag_construction_error(cls, message, exc=None):
        if exc is not None:
            raise cls.__get_exception(message) from exc
        else:
            raise cls.__get_exception(message)

    @classmethod
    def map_node_data(cls, context, *args, **kwargs):
        raise NotImplementedError()

    @classmethod
    def from_yaml(cls, loader, node):
        if type(node) is ScalarNode:
            loaded_node = loader.construct_scalar(node)
            return cls.map_node_data(loader.context, loaded_node)
        if type(node) is SequenceNode:
            loaded_node = loader.construct_sequence(node, deep=True)
            return cls.map_node_data(loader.context, *loaded_node)
        if type(node) is MappingNode:
            loaded_node = loader.construct_mapping(node, deep=True)
            return cls.map_node_data(loader.context, **loaded_node)


class EnvVar(ArgListAcceptingYamlTag):
    yaml_tag = "!EnvVar"

    @classmethod
    def map_node_data(cls, context, name):
        if name not in os.environ:
            cls.handle_tag_construction_error("Environment Variable '{}' not set!".format(name))

        return os.environ[name]


class RelativeFileIncludingYamlTag(ArgListAcceptingYamlTag):
    @classmethod
    def handle_included_file(cls, file_handle):
        raise NotImplementedError()

    @classmethod
    def map_node_data(cls, context, filename):
        current_file_directory = os.path.dirname(context["filename"])
        full_file_path = os.path.join(current_file_directory, filename)
        try:
            with open(full_file_path) as file_handle:
                return cls.handle_included_file(file_handle)

        except IOError as e:
            cls.handle_tag_construction_error(
                "Attempted to include relative file {}, which doesn't exist!".format(filename),
                e,
            )


class IncludeJson(RelativeFileIncludingYamlTag):
    yaml_tag = "!IncludeJson"

    @classmethod
    def handle_included_file(cls, file_handle):
        try:
            return json.load(file_handle)
        except ValueError as e:
            cls.handle_tag_construction_error("Failed to parse relative json file {}!".format(filename), e)


class IncludeYaml(RelativeFileIncludingYamlTag):
    yaml_tag = "!IncludeYaml"

    @classmethod
    def handle_included_file(cls, file_handle):
        try:
            context_passing_loader = lambda stream: ContextPassingYamlLoader(stream, {"filename": file_handle.name})
            return load(file_handle, Loader=context_passing_loader)
        except ValueError as e:
            cls.handle_tag_construction_error("Failed to parse relative yaml file {}!".format(filename), e)


class IncludeText(RelativeFileIncludingYamlTag):
    yaml_tag = "!IncludeText"

    @classmethod
    def handle_included_file(cls, file_handle):
        return file_handle.read().strip()



#!/usr/bin/env python


class FileParsingException(Exception):
    def __init__(self, filename):
        super().__init__("Exception thrown while processing {}".format(filename))


class FilesNotFoundException(Exception):
    def __init__(self, msg):
        super().__init__(msg)



#!/usr/bin/env python
import logging
import logging.config
import os

from .parsers import SUPPORTED_FILE_EXTENSIONS, parse_file
from .exceptions import FilesNotFoundException

__CONFIG_FILENAME_FORMAT = "{basename}{extension}"
__LOGGER = logging.getLogger(__name__)


def __get_configuration_dirs(configuration_dirs_arg):
    if configuration_dirs_arg is None:
        return [os.path.join(os.getcwd(), "config")]

    elif type(configuration_dirs_arg) is basestring:
        return [configuration_dirs_arg]

    return configuration_dirs_arg


def __merge_configuration_from_dict_root(base_config, overrides):
    for k, v in overrides.items():
        if type(v) is dict and type(base_config.get(k)) is dict:
            __merge_configuration_from_dict_root(base_config[k], v)
        else:
            base_config[k] = v


def __merge_configuration_from_file(base_config, filename, fail_on_parse_error):
    overrides = parse_file(filename, fail_on_parse_error)
    __merge_configuration_from_dict_root(base_config, overrides)


def __find_available_config_files_in_directory(directory, basename):
    possible_config_files = (
        os.path.join(directory, __CONFIG_FILENAME_FORMAT.format(basename=basename, extension=extension))
        for extension in SUPPORTED_FILE_EXTENSIONS
    )

    return [config_file for config_file in possible_config_files if os.path.isfile(config_file)]


def __handle_available_files_in_directories(
    base_config,
    file_basenames,
    configuration_dirs,
    fail_on_parse_error,
    fail_on_missing_files
):
    basename_found = {b: False for b in file_basenames}

    for directory in configuration_dirs:
        for basename in file_basenames:
            config_files_in_dir = __find_available_config_files_in_directory(directory, basename)

            if len(config_files_in_dir) == 0:
                __LOGGER.info("No config files for basename {} found in directory {}".format(basename, directory))
            else:
                basename_found[basename] = True

            for f in config_files_in_dir:
                __LOGGER.info("Parsing file {} and merging with config".format(f))
                __merge_configuration_from_file(
                    base_config=base_config,
                    filename=f,
                    fail_on_parse_error=fail_on_parse_error,
                )

    for basename, found in basename_found.items():
        if fail_on_missing_files and not found:
            __LOGGER.error("No files found for basename {} in any directory and fail_on_missing_files is set, exiting".format(basename))
            raise FilesNotFoundException("No files found for basename {} in any directory".format(basename))


def __handle_available_defaults_files(
    base_config,
    defaults_basename,
    configuration_dirs,
    fail_on_parse_error,
    fail_on_missing_files
):
    __LOGGER.info("Searching for defaults files...")

    __handle_available_files_in_directories(
        base_config=base_config,
        file_basenames=[defaults_basename],
        configuration_dirs=configuration_dirs,
        fail_on_parse_error=fail_on_parse_error,
        fail_on_missing_files=fail_on_missing_files,
    )


def __handle_active_profiles_files(
    base_config,
    active_profiles,
    configuration_dirs,
    fail_on_parse_error,
    fail_on_missing_files
):
    __LOGGER.info("Searching for active profile files...")

    __handle_available_files_in_directories(
        base_config=base_config,
        file_basenames=active_profiles,
        configuration_dirs=configuration_dirs,
        fail_on_parse_error=fail_on_parse_error,
        fail_on_missing_files=fail_on_missing_files,
    )


def __configure_logging(
    configuration_dirs,
    logging_config_filename,
    fail_on_parse_error,
    fail_on_missing_files
):
    logging_config = {}

    __handle_available_files_in_directories(
        base_config=logging_config,
        file_basenames=[logging_config_filename],
        configuration_dirs=configuration_dirs,
        fail_on_parse_error=fail_on_parse_error,
        fail_on_missing_files=fail_on_missing_files,
    )

    logging.config.dictConfig(logging_config)
    __LOGGER.info("Configured logging")


def configure(
    configuration_dirs=None,
    logging_config_filename="logging",
    defaults_basename="defaults",
    active_profiles=None,
    fail_on_parse_error=True,
    fail_on_missing_files=False,
):
    """
    :param configuration_dirs: The directories from which configuration files will be pulled, either a single string, or
                               a list of strings. If None is passed, defaults to a subdirectory of the cwd called
                               "config"
    :param logging_config_filename: If provided (not None), read config files matching this base name and any supported
                                    extension and use the read config to configure the __LOGGER
    :param defaults_basename: The name of the default config files that are read first, same as the profile, it is the
                              basename of the file to look for without an extension. Defaults to "defaults"
    :param active_profiles: The list of profiles currently active, if None is passed, this will be read from an
                            environment variable JCONFIGURE_ACTIVE_PROFILES. This variable should be a comma separated
                            list of strings. For each profile a file with the name {profile}.{extension} will be read,
                            where extension is one of the allowed file types
    :param fail_on_parse_error: If False suppress any exceptions thrown while processing a file. Defaults to True
    :param fail_on_missing_files: If True, raise an exception if an expected file is not found. Defaults to False

    :return: The configuration dictionary pulled from the configuration files specified under configuration_dir
    """
    configuration_dirs = __get_configuration_dirs(configuration_dirs)

    __configure_logging(
        configuration_dirs=configuration_dirs,
        logging_config_filename=logging_config_filename,
        fail_on_parse_error=fail_on_parse_error,
        fail_on_missing_files=fail_on_missing_files,
    )

    active_profiles = (
        active_profiles or
        os.environ.get("JCONFIGURE_ACTIVE_PROFILES").split(",") if "JCONFIGURE_ACTIVE_PROFILES" in os.environ else []
    )

    base_config = {}

    __LOGGER.info("Configuring Application using files in config directories [{}]".format(", ".join(configuration_dirs)))
    __LOGGER.info("Active profiles: [{}]".format(", ".join(active_profiles)))

    __handle_available_defaults_files(
        base_config=base_config,
        configuration_dirs=configuration_dirs,
        defaults_basename=defaults_basename,
        fail_on_parse_error=fail_on_parse_error,
        fail_on_missing_files=fail_on_missing_files,
    )

    __handle_active_profiles_files(
        base_config=base_config,
        configuration_dirs=configuration_dirs,
        active_profiles=active_profiles,
        fail_on_parse_error=fail_on_parse_error,
        fail_on_missing_files=fail_on_missing_files,
    )

    return base_config



#!/usr/bin/env python
import json
import logging
import os
import yaml

from .exceptions import FileParsingException
from .yaml_tags import *

__LOGGER = logging.getLogger(__name__)


class JsonConfigFileParser:
    FILE_EXTENSIONS = [".json"]

    @staticmethod
    def parse(filename):
        with open(filename) as json_file:
            return json.load(json_file)


class YamlConfigFileParser:
    FILE_EXTENSIONS = [".yaml", ".yml"]

    @staticmethod
    def parse(filename):
        with open(filename) as yaml_file:
            context_passing_loader = lambda stream: ContextPassingYamlLoader(stream, {"filename": filename})
            return yaml.load(yaml_file, Loader=context_passing_loader)


__AVAILABLE_FILE_PARSERS = [
    JsonConfigFileParser,
    YamlConfigFileParser,
]

__FILE_EXTENSION_TO_PARSERS = {
    extension: parser
    for parser in __AVAILABLE_FILE_PARSERS
    for extension in parser.FILE_EXTENSIONS
}

SUPPORTED_FILE_EXTENSIONS = list(__FILE_EXTENSION_TO_PARSERS.keys())


def parse_file(filename, fail_on_parse_error=True):
    _, extension = os.path.splitext(filename)
    parser = __FILE_EXTENSION_TO_PARSERS[extension]

    try:
        return parser.parse(filename)
    except Exception as e:
        if fail_on_parse_error:
            __LOGGER.error("Exception thrown while parsing file {} using parser {}!".format(filename, parser))
            raise FileParsingException(filename) from e
        else:
            __LOGGER.warn("Exception thrown while parsing file {} using parser {}. fail_on_parse_error is not set, continuing".format(
                filename, parser
            ))

            return {}
